/*----------------------------------------------------------------------------*
 | Headerfile CallbackR0.h                                                    |
 |                                                                            |
 | Version 2.00                                                               |
 |                                                                            |
 |                                                                            |
 | BriDGe::Callback: a tiny library that provides callback functionality.     |
 | Copyright (C) 2002  Bram de Greve                                          |
 |                                                                            |
 | This library is free software; you can redistribute it and/or              |
 | modify it under the terms of the GNU Lesser General Public                 |
 | License as published by the Free Software Foundation; either               |
 | version 2.1 of the License, or (at your option) any later version.         |
 |                                                                            |
 | This library is distributed in the hope that it will be useful,            |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of             |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU          |
 | Lesser General Public License for more details.                            |
 |                                                                            |
 | You should have received a copy of the GNU Lesser General Public           |
 | License along with this library; if not, write to the Free Software        |
 | Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  |
 |                                                                            |
 |                                                                            |
 | Contact by e-mail: Bram de Greve <bdegreve@mail.be>                        |
 |                                                                            |
 |                                                                            |
 | History:                                                                   |
 | May 28th, 2002: creation of CallbackR0.h                                   |
 | June 2th, 2002: Getting rid of interal compiler error.                     |
 |                 Yiha!  We've bypassed the system!  mwuahaha!  hihi :)      |
 | June 7th, 2002: Splitting CallbackR0 to multiple files.                    |
 | June 8th, 2002: Revamping the code.                                        |
 |                 Putting stuff under GNU LGPL.                              |
 | June 9th, 2002: Getting rid of boost::shared_ptr so we look less silly,    |
 |                 Getting rid of makeCallback since we use templated         |
 |                 constructors now.                                          |
 |                 Making the call of an empty callback a little more safe    |
 |                 by throwing an exception CallbackEmptyCall.                |
 |                 Getting rid of operator bool and renaming isEmpty to empty.|
 |                 And some other stuff too.  Thank you Dirk Gerrits!         |
 |                 Soit, we have now officially reached version 2.00!         |
 |                                                                            |
 | I want to thank Altair, Jaap Suter & especially Dirk Gerrits (Who has      |
 | spent an entire afternoon on this to help me.  It was a waste of time for  |
 | both of us, but we had a lot of fun, didn't we? :)                         |
 *----------------------------------------------------------------------------*/

// HOW TO USE THIS:
//
// 1. You create a CallbackR0 by using a ... constructor.  Surprise, surprise!
//    It used to be done by maker functions, but we have now templated
//    constructors.  Very cool:
//          int foo();
//          CallbackR0<int> callback1(foo);
//
//          class Bar
//          {
//          public:
//              int fun();
//          }
//          Bar bar;
//          CallbackR0<int> callback2(bar, Bar::fun);
// 
// 2. You can copy callback objects as much as you like.
//          CallbackR0<int> callback3(callback1);
//
//          CallbackR0<int> callback4;
//          callback4 = callback2;
//
// 3. To excecute the callback, you evoke the function operator
//          int r = callback1();
//
// Note: Although you can allocate callbacks without specifying the exact
//       callback function, you must make sure the callback can do something
//       when you excecute it.  So, you must feed it with appropriate beef 
//       (functions, object/method pair).  If you don't, funny things will
//       happen.  A crash for example.  Luckely for you, a special exception
//       BDG::CallbackEmptyCall will be thrown so that you can catch it.
//       On the other hand, you could also have checked this before the call
//       with the method empty().
//          try
//          {
//              int r = callback1();
//          }
//          catch (CallbackEmptyCall)
//          {
//              ...
//          }
//
//          if (!callback1.empty())
//          {
//              int r = callback1();
//          }
//
// Note: If you use a object/method pair as callback, you must make sure the 
//       object is alive when calling the callback.  Otherwise the callback
//       will want to do the method on dead beef, and you know funny things
//       happen when it does that.  And you know what that means, don't you?



#ifndef BDG_GUARDIAN_OF_INCLUSION_CALLBACKR0_H
#define BDG_GUARDIAN_OF_INCLUSION_CALLBACKR0_H



// --- OLD INTERFACES ----------------------------------------------------------

#include "CallbackR0Detail.h"
#include "CallbackExceptions.h"
#include <algorithm>                    // std::swap



// --- NEW INTERFACES ----------------------------------------------------------

namespace BDG
{


// THE CALLBACK:
// This is the actuall class you use to hold callbacks:

template <typename R>
class CallbackR0
{
public:

    // CONSTRUCTORS & DESTRUCTORS
    
    CallbackR0(): 
        m_dispatcher(0) 
    {
    }

    template<typename FR>
    explicit CallbackR0(FR (*a_function)())
    {
        if (!a_function) throw CallbackNullArgument(
            "BDG::CallbackR0: tried to construct from null function pointer.");
        using namespace CallbackR0_Detail;
        m_dispatcher = new DispFunction<R, FR>(a_function);
    }
    
    template<class Object, typename MR>
    CallbackR0(Object* a_object, MR (Object::*a_method)())
    {
        if (!(a_object && a_method)) throw CallbackNullArgument(
            "BDG::CallbackR0: tried to construct from null object or method "
            "pointer.");
        using namespace CallbackR0_Detail;
        m_dispatcher = new DispMethod<Object, R, MR>(a_object, a_method);
    }

    template<class Object, typename MR>
    CallbackR0(Object* a_object, MR (Object::*a_method)() const)
    {
        if (!(a_object && a_method)) throw CallbackNullArgument(
            "BDG::CallbackR0: tried to construct from null object or method "
            "pointer.");
        using namespace CallbackR0_Detail;
        m_dispatcher = new DispMethodConst<Object, R, MR>(a_object, a_method);
    }

    CallbackR0(const CallbackR0<R>& a_other):
        m_dispatcher(a_other.empty() ? 0 : a_other.m_dispatcher->clone())
    {
    }
    
    
    // OPERATORS

    CallbackR0 operator=(const CallbackR0<R>& a_other)
    {
        CallbackR0<R> temp(a_other);
        swap(temp);
        return *this;
    }           
    
    R operator()() const 
    {
        if (empty()) throw CallbackEmptyCall(
            "BDG::CallbackR0: tried to call an empty callback.");
        return (*m_dispatcher)(); 
    }


    // METHODS

    bool empty() const
    {
        return m_dispatcher == 0;
    }

    void swap(CallbackR0<R>& a_other)
    {
        std::swap(m_dispatcher, a_other.m_dispatcher);
    }

private:

    CallbackR0_Detail::DispBase<R>* m_dispatcher;
};



} // namespace BDG

#endif // Guardian of Inclusion

// --- END OF FILE CallbackR0.h ------------------------------------------------
